# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import importlib
import os

top = os.path.dirname(__file__)
for _, dirnames, filenames in os.walk(top):
    dirnames[:] = []
    for filename in filenames:
        if filename.startswith("-"):
            continue
        importlib.import_module(
            f".{os.path.splitext(filename)[0]}", "cxx_flow.commands"
        )



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

from abc import ABC, abstractmethod
from typing import List

from .config import Config, Runtime


class Step(ABC):
    @property
    @abstractmethod
    def name(self) -> str: ...

    @property
    def runs_after(self) -> List[str]:
        return []

    def platform_dependencies(self) -> List[str]:
        return []

    def is_active(self, config: Config, rt: Runtime) -> bool:
        return True

    def directories_to_remove(self, config: Config) -> List[str]:
        return []

    @abstractmethod
    def run(self, config: Config, rt: Runtime) -> int: ...


__steps: List[Step] = []


def register_step(step: Step):
    global __steps

    name = step.name
    if name in [step.name for step in __steps]:
        raise NameError(f"Step {name} already registered")

    __steps.append(step)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import pprint
import re

import chevron

from cxx_flow.flow import config, ctx, init

from .__version__ import CMAKE_VERSION

config_json_mustache = """
{{#with_cmake}}
    "cmake": {
        "vars": {
            "{{NAME_PREFIX}}_COVERAGE": "?config:coverage",
            "{{NAME_PREFIX}}_SANITIZE": "?config:sanitizer",
            "{{NAME_PREFIX}}_CUTDOWN_OS": "?runtime:cutdown_os"
        }
    },
{{/with_cmake}}
"""


class CMakeInit(init.InitStep):
    def postprocess(self, rt: config.Runtime, context: dict):
        patch = chevron.render(config_json_mustache, context).rstrip()
        if not patch:
            return

        with open(".flow/config.json", encoding="UTF-8") as config_file:
            patched = re.split(r'(\s*"shortcuts":\s*{)', config_file.read())

        if len(patched) != 3:
            return

        patched.insert(1, patch)
        content = "".join(patched)

        with open(".flow/config.json", "w", encoding="UTF-8") as config_file:
            config_file.write(content)


def _list_cmake_types():
    return ctx.move_to_front(
        "console-application",
        sorted(key for key in ctx.get_internal("cmake").keys() if key),
    )


init.register_init_step(CMakeInit())
ctx.register_init_setting(
    ctx.Setting("PROJECT.TYPE", "CMake project type", _list_cmake_types)
)
ctx.register_init_setting(
    ctx.Setting("cmake", fix="{PROJECT.TYPE$map:cmake}"),
    ctx.Setting("CMAKE_VERSION", value=CMAKE_VERSION),
    is_hidden=True,
)
ctx.register_switch("with_cmake", "Use CMake", True)
ctx.register_internal(
    "cmake",
    {
        "": {"cmd": "add_executable", "type": ""},
        "console-application": {
            "cmd": "add_executable",
            "type": "",
            "console-application": True,
            "console": True,
            "application": True,
            "link_access": "PRIVATE",
        },
        "win32-application": {
            "cmd": "add_executable",
            "type": " WIN32",
            "win32-application": True,
            "win32": True,
            "application": True,
            "link_access": "PRIVATE",
        },
        "macos-application": {
            "cmd": "add_executable",
            "type": " MACOSX_BUNDLE",
            "macos-application": True,
            "macos": True,
            "bundle": True,
            "application": True,
            "link_access": "PRIVATE",
        },
        "static-library": {
            "cmd": "add_library",
            "type": " STATIC",
            "static-library": True,
            "static": True,
            "library": True,
            "link_access": "PUBLIC",
        },
        "shared-library": {
            "cmd": "add_library",
            "type": " SHARED",
            "shared-library": True,
            "shared": True,
            "library": True,
            "link_access": "PUBLIC",
        },
        "plugin-library": {
            "cmd": "add_library",
            "type": " MODULE",
            "plugin-library": True,
            "plugin": True,
            "library": True,
            "link_access": "PUBLIC",
        },
    },
)



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import os

from cxx_flow.flow.config import Config, Runtime
from cxx_flow.flow.step import Step, register_step

from .__version__ import CMAKE_VERSION


class CMakeBuild(Step):
    name = "Build"
    runs_after = ["Conan", "CMake"]

    def platform_dependencies(self):
        return [f"cmake>={CMAKE_VERSION}"]

    def is_active(self, config: Config, rt: Runtime) -> int:
        return os.path.isfile("CMakeLists.txt") and os.path.isfile("CMakePresets.json")

    def run(self, config: Config, rt: Runtime) -> int:
        rt.cmd("cmake", "--build", "--preset", config.preset, "--parallel")


register_step(CMakeBuild())



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import json
import os
import shutil
from dataclasses import dataclass
from typing import List, Optional, cast

import chevron

from ..flow import ctx
from ..flow.config import Runtime


@dataclass
class FileInfo:
    src: str
    dst: str
    is_mustache: bool
    when: Optional[str] = None

    @classmethod
    def from_json(cls, src: str, filelist: dict, context: ctx.SettingsType):
        basename, ext = os.path.splitext(src)
        is_mustache = ext == ".mustache"
        json_file = cast(dict, filelist.get(src, {}))
        path = cast(Optional[str], json_file.get("path"))
        when = cast(Optional[str], json_file.get("when"))
        dst = (
            chevron.render(path, context)
            if path is not None
            else basename if is_mustache else src
        )
        return cls(src=src, dst=dst, is_mustache=is_mustache, when=when)

    def template(self):
        open_mstch = "{{"
        close_mstch = "}}"
        if self.when:
            return f"{open_mstch}#{self.when}{close_mstch}\n{self.dst}\n{open_mstch}/{self.when}{close_mstch}\n"
        return f"{self.dst}\n"

    def run(self, root: str, rt: Runtime, context: ctx.SettingsType):
        if not rt.silent:
            if rt.use_color:
                print(f"\033[2;30m+\033[m {self.dst}")
            else:
                print(f"+ {self.dst}")
        if rt.dry_run:
            return

        src = os.path.join(root, self.src)
        dst = os.path.abspath(self.dst)
        dirname = os.path.dirname(dst)
        os.makedirs(dirname, exist_ok=True)

        if self.is_mustache:
            with open(src, "r", encoding="UTF-8") as inf:
                content = inf.read()
            content = chevron.render(content, context)
            with open(dst, "w", encoding="UTF-8") as outf:
                outf.write(content)
        else:
            shutil.copy(src, dst, follow_symlinks=False)


@dataclass
class LayerInfo:
    root: str
    files: List[FileInfo]
    when: Optional[str] = None

    @classmethod
    def from_fs(cls, layer_dir: str, context: ctx.SettingsType):
        with open(f"{layer_dir}.json", encoding="UTF-8") as f:
            layer_info: dict = json.load(f)
        when = cast(Optional[bool], layer_info.get("when"))
        filelist = cast(dict, layer_info.get("filelist", {}))

        sources: List[str] = []
        prefix = os.path.join(layer_dir, "")
        for curr_dir, dirs, files in os.walk(layer_dir):
            dirs[:] = [dirname for dirname in dirs if dirname not in ["__pycache__"]]
            files[:] = [
                filename
                for filename in files
                if os.path.splitext(filename) not in [".pyc", ".pyo", ".pyd"]
            ]

            for filename in files:
                src = os.path.join(curr_dir, filename)
                if src.startswith(prefix):
                    src = src[len(prefix) :]
                sources.append(src)

        files = [FileInfo.from_json(src, filelist, context) for src in sources]
        files.sort(key=lambda info: info.dst)

        result = cls(root=layer_dir, files=files, when=when)
        allowed_files = set(
            filter(
                lambda path: path != "",
                chevron.render(result.template(), context).split("\n"),
            )
        )
        result.files = list(filter(lambda file: file.dst in allowed_files, files))

        return result

    @property
    def name(self):
        return os.path.basename(self.root)

    @property
    def pkg(self):
        return os.path.basename(
            os.path.dirname(
                os.path.dirname(os.path.dirname(os.path.dirname(self.root)))
            )
        )

    def template(self):
        open_mstch = "{{"
        close_mstch = "}}"
        result = ""
        if self.when:
            result += f"{open_mstch}#{self.when}{close_mstch}\n"
        result += "".join(file.template() for file in self.files)
        if self.when:
            result += f"{open_mstch}/{self.when}{close_mstch}\n"
        return result

    def run(self, rt: Runtime, context: ctx.SettingsType):
        if not rt.silent:
            if rt.use_color:
                print(f"\033[2;30m[{self.pkg}:{self.name}]\033[m")
            else:
                print(f"[{self.pkg}:{self.name}]")
        for file in self.files:
            file.run(self.root, rt, context)
        if not rt.silent:
            print()


def copy_license(rt: Runtime, context: ctx.SettingsType):
    license = context.get("COPY", {}).get("LICENSE")
    if not license:
        return

    licenses_dir = os.path.abspath(
        os.path.join(ctx.package_root, ctx.template_dir, "licenses")
    )
    license_file = f"{license}.mustache"
    info = FileInfo(license_file, "LICENSE", is_mustache=True)
    info.run(licenses_dir, rt, context)


def gather_package_layers(package_root: str, context: ctx.SettingsType):
    layers_dir = os.path.abspath(os.path.join(package_root, ctx.template_dir, "layers"))
    for root, dirs, files in os.walk(layers_dir):
        layer_dirs = [
            os.path.abspath(os.path.join(root, dirname))
            for dirname in dirs
            if f"{dirname}.json" in files
        ]
        dirs[:] = []

    layers = (LayerInfo.from_fs(layer_dir, context) for layer_dir in layer_dirs)
    return list(filter(lambda layer: len(layer.files) > 0, layers))



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)


import argparse
import json
import os
import shlex
import subprocess
import sys
from dataclasses import dataclass
from typing import Dict, List, Optional, cast

from .matrix import cartesian, find_compiler, flatten, load_matrix, matches_any
from .uname import uname

platform = uname()[0]


def _compiler_inner(
    value: str, used_compilers: Dict[str, List[str]], config_names: Dict[str, List[str]]
):
    compiler, names = find_compiler(value, config_names)
    if compiler not in used_compilers:
        used_compilers[compiler] = []
    used_compilers[compiler].append(names)
    return compiler


def _compiler(used_compilers: Dict[str, List[str]], config_names: Dict[str, List[str]]):
    return lambda value: _compiler_inner(value, used_compilers, config_names)


def _boolean_inner(value: str, with_name: str):
    v = value.lower()
    return v in _TRUE or v == with_name


def _boolean(with_name: str):
    return lambda value: _boolean_inner(value, with_name)


_TRUE = {"true", "on", "yes", "1"}
boolean_sanitizer = _boolean("with-sanitizer")


def _types(used_compilers: Dict[str, List[str]], config_names: Dict[str, List[str]]):
    return {
        "compiler": _compiler(used_compilers, config_names),
        "sanitizer": boolean_sanitizer,
    }


def _config(config: List[str], only_host: bool, types: Dict[str, callable]):
    args = {}
    for arg in config:
        if arg[:1] == "-":
            continue
        _arg = arg.split("=", 1)
        if len(_arg) == 1:
            continue

        name, vals = _arg
        name = name.strip()
        conv = types.get(name, lambda value: value)
        values = {conv(val.strip()) for val in vals.split(",")}
        if name in args:
            values.update(args[name])
        args[name] = list(values)

    if only_host and "os" not in args:
        args["os"] = [platform]

    return args


def _expand_one(config: dict, github_os: str, os_in_name: str):
    os_ver = github_os.split("-")[1]
    build_name = f"{config['build_type']} with {config['compiler']} on {os_in_name}"
    if config["sanitizer"]:
        build_name += " (and sanitizer)"
    config["github_os"] = github_os
    config["build_name"] = build_name
    config["needs_gcc_ppa"] = (
        os_ver != "latest"
        and config["os"] == "ubuntu"
        and int(os_ver.split(".")[0]) < 24
    )
    return config


def _expand_config(config: dict, spread_lts: bool, lts_list: Dict[str, List[str]]):
    if spread_lts:
        spread = lts_list.get(config["os"], [])
        if len(spread):
            return [
                _expand_one({key: config[key] for key in config}, lts, lts)
                for lts in spread
            ]
    return [_expand_one(config, f"{config['os']}-latest", config["os"])]


_flow_config_default_compiler: Optional[Dict[str, str]] = None


def default_compiler():
    try:
        return os.environ["DEV_CXX"]
    except KeyError:
        pass

    try:
        return _flow_config_default_compiler[platform]
    except KeyError:
        print(f"KeyError: {platform} in {_flow_config_default_compiler}")
        return "?"
    except TypeError:
        print(f"TypeError: internal: flow config not ready yet")
        return "?"


def _load_flow_data(rt: "Runtime"):
    root = ".flow"
    paths = [os.path.join(root, "matrix.json")]
    if rt.official:
        paths.append(os.path.join(root, "official.json"))
    matrix, keys = load_matrix(*paths)

    if rt.no_coverage:
        for conf in matrix:
            if "coverage" in conf:
                del conf["coverage"]

    return matrix, keys


def _hide(arg: str, secrets: List[str]):
    for secret in secrets:
        arg = arg.replace(secret, "?" * max(15, len(secret)))
    return arg


def _print_arg(arg: str, secrets: List[str]):
    color = ""
    arg = _hide(arg, secrets)
    if arg[:1] == "-":
        color = "\033[2;37m"
    arg = shlex.join([arg])
    if color == "" and arg[:1] in ["'", '"']:
        color = "\033[2;34m"
    if color == "":
        return arg
    return f"{color}{arg}\033[m"


def _print_cmd(*args: str, use_color: bool = True, secrets: List[str]):
    if not use_color:
        cmd = shlex.join([args[0]])
        print(cmd, shlex.join(_hide(arg) for arg in args[1:]))
        return

    cmd = shlex.join([args[0]])
    args = " ".join([_print_arg(arg, secrets) for arg in args[1:]])
    print(f"\033[33m{cmd}\033[m {args}", file=sys.stderr)


@dataclass
class RunAlias:
    name: str
    doc: str
    steps: List[str]

    @staticmethod
    def from_json(name: str, alias: dict):
        doc: str = alias.get("doc", "")
        steps: List[str] = alias.get("steps", [])
        if not doc:
            doc = f'shortcut for "run -s {",".join(steps)}"'

        return RunAlias(name, doc, steps)


class FlowConfig:
    _cfg: dict
    steps: list = []
    aliases: List[RunAlias] = []

    def __init__(self):
        global _flow_config_default_compiler

        try:
            with open(
                os.path.join(".flow", "config.json"),
                encoding="UTF-8",
            ) as f:
                self._cfg = json.load(f)
        except FileNotFoundError:
            self._cfg = {}

        _flow_config_default_compiler = self.compiler_os_default

    @property
    def entry(self) -> Dict[str, dict]:
        return self._cfg.get("entry", {})

    @property
    def compiler(self) -> Dict[str, dict]:
        return self._cfg.get("compiler", {})

    @property
    def compiler_names(self) -> Dict[str, List[str]]:
        return self.compiler.get("names", {})

    @property
    def compiler_os_default(self) -> Dict[str, str]:
        return self.compiler.get("os-default", {})

    @property
    def lts_list(self) -> Dict[str, List[str]]:
        return self._cfg.get("lts", {})

    @property
    def postproc(self) -> dict:
        return self._cfg.get("postproc", {})

    @property
    def postproc_exclude(self) -> List[dict]:
        return self.postproc.get("exclude", [])

    @property
    def compiler_names(self) -> Dict[str, List[str]]:
        return self.compiler.get("names", {})

    @property
    def shortcuts(self) -> Dict[str, dict]:
        return self._cfg.get("shortcuts", {})


class Runtime(FlowConfig):
    dry_run: bool
    silent: bool
    official: bool
    no_coverage: bool
    use_color: bool
    platform: str
    secrets: List[str] = []

    def __init__(self, args: argparse.Namespace):
        super().__init__()

        self.dry_run = args.dry_run
        self.silent = args.silent
        try:
            self.official = args.official
        except AttributeError:
            self.official = False
        self.use_color = True
        self.no_coverage = False
        self.platform = platform

        if "NO_COVERAGE" in os.environ:
            self.no_coverage = True

        if "RELEASE" in os.environ and "GITHUB_ACTIONS" in os.environ:
            self.official = not not json.loads(os.environ["RELEASE"])

    @property
    def only_host(self):
        return not (self.dry_run or self.official)

    def cmd(self, *args: str):
        if not self.silent:
            _print_cmd(*args, use_color=self.use_color, secrets=self.secrets)
        if not self.dry_run:
            result = subprocess.run(args)
            if result.returncode != 0:
                print(
                    f"cxx-flow: error: {args[0]} ended in failure, exiting",
                    file=sys.stderr,
                )
                sys.exit(1)


@dataclass
class Config:
    items: dict
    keys: List[str]

    def get_path(self, key: str, val: any = None):
        path = key.split(".")
        ctx = self.items
        for step in path:
            if not isinstance(ctx, dict):
                return val
            child = ctx.get(step)
            if child is None:
                return val
            ctx = child
        return cast(any, ctx)

    @property
    def build_type(self) -> str:
        return self.items.get("build_type", "")

    @property
    def build_name(self) -> str:
        return self.items.get("build_name", "")

    @property
    def preset(self) -> str:
        return self.items.get("preset", "")

    @property
    def build_generator(self) -> str:
        return self.items.get("build_generator", "")


class Configs:
    usable: List[Config] = []

    def __init__(self, args: argparse.Namespace):
        super()
        rt = Runtime(args)
        matrix, keys = _load_flow_data(rt)

        used_compilers: Dict[str, List[str]] = {}

        types = _types(used_compilers=used_compilers, config_names=rt.compiler_names)
        arg_configs = cartesian(_config(flatten(args.configs), rt.only_host, types))

        # from commands/github
        spread_lts = hasattr(args, "matrix") and not not args.matrix

        turned = flatten(
            [
                _expand_config(config, spread_lts, rt.lts_list)
                for config in matrix
                if len(arg_configs) == 0 or matches_any(config, arg_configs)
            ]
        )

        postproc_exclude = rt.postproc_exclude
        usable = [
            config
            for config in turned
            if len(postproc_exclude) == 0 or not matches_any(config, postproc_exclude)
        ]

        self.usable = []
        for conf in usable:
            try:
                compilers = used_compilers[conf["compiler"]]
            except KeyError:
                fallback_compiler = find_compiler(
                    conf["compiler"], config_names=rt.compiler_names
                )
                compilers = [fallback_compiler[1]]
            for compiler in compilers:
                self.usable.append(
                    Config(
                        {
                            **conf,
                            "compiler": compiler,
                            "--orig-compiler": conf["compiler"],
                        },
                        keys,
                    )
                )



# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

import importlib
import os
from dataclasses import dataclass
from types import ModuleType
from typing import List, Union

from . import step
from .config import FlowConfig, RunAlias


@dataclass
class Sorted:
    plugin: step.Step
    name: str
    runs_after: List[str]

    @staticmethod
    def from_step(plugin: step.Step):
        name = plugin.name
        runs_after = [*plugin.runs_after]
        return Sorted(plugin, name, runs_after)


def _load_plugins(directory: str, package: Union[str, None]):
    for _, dirnames, filenames in os.walk(directory):
        for dirname in dirnames:
            importlib.import_module(f".{dirname}", package=package)
        for filename in filenames:
            if filename == "__init__.py":
                continue
            importlib.import_module(
                f".{os.path.splitext(filename)[0]}", package=package
            )
        dirnames[:] = []
        pass


def _load_module_plugins(cfg: FlowConfig, mod: ModuleType):
    spec = mod.__spec__
    if not spec:
        return
    for location in spec.submodule_search_locations:
        _load_plugins(location, spec.name)


def _sort_steps():
    steps = step.__steps
    unsorted = [Sorted.from_step(step) for step in steps]
    known_names = [step.name for step in unsorted]

    for plugin in unsorted:
        runs_after: List[str] = []
        for name in plugin.runs_after:
            if name in known_names:
                runs_after.append(name)
        plugin.runs_after = runs_after

    result: List[step.Step] = []

    while len(unsorted) > 0:
        layer = [plugin for plugin in unsorted if len(plugin.runs_after) == 0]
        unsorted = [plugin for plugin in unsorted if len(plugin.runs_after) > 0]
        for plugin in layer:
            for remaining in unsorted:
                try:
                    remaining.runs_after.remove(plugin.name)
                except ValueError:
                    pass
        result.extend(plugin.plugin for plugin in layer)
        if len(layer) == 0:
            result.extend(plugin.plugin for plugin in unsorted)
            break

    return result


def load_steps(cfg: FlowConfig):
    std_plugins = importlib.import_module("cxx_flow.plugins")
    _load_module_plugins(cfg, std_plugins)
    # TODO: use cfg.steps.plugins

    return _sort_steps()


def clean_aliases(cfg: FlowConfig, valid_steps: List[step.Step]):
    entries = cfg._cfg.get("entry")
    if not entries:
        return

    step_names = {step.name for step in valid_steps}

    keys_to_remove: List[str] = []
    for key in entries:
        entry = entries[key]
        steps = entry.get("steps")
        if not steps:
            keys_to_remove.append(key)
            continue
        rewritten: List[str] = []
        for step in steps:
            if step in step_names:
                rewritten.append(step)
        steps[:] = rewritten

        if not steps:
            keys_to_remove.append(key)

    for key in keys_to_remove:
        del entries[key]

    cfg.steps = valid_steps
    cfg.aliases = [RunAlias.from_json(key, entries[key]) for key in entries]
